module spi_gen (
  input clk,
  input rst,

	output reg spi_cs,
	output reg spi_sdo,
  output reg spi_sdo_d,
	inout      spi_sdi,
	output reg spi_sck,
  output reg spi_sdi_as_output
);

localparam WORDS_DELAY = 10;
localparam MAGIC_TO_DATA_DELAY = 50;
localparam SPI_WORDS       = 22;
localparam SPI_WORDS_MAGIC = 6;

localparam DATA_W = 16;

localparam MAGIC_WORD = 16'hAABB;
localparam MAGIC_RD   = 16'h0001;
localparam MAGIC_WR   = 16'h0000;
localparam MAGIC_CPU_RST   = 16'h0002;
localparam MAGIC_CPU_UNRST = 16'h0003;

reg [DATA_W-1:0] spi_data [SPI_WORDS - 1:0];

initial begin : ram_fill
  spi_data[ 0] = MAGIC_WORD;
  spi_data[ 1] = MAGIC_RD;//MAGIC_WR;
  spi_data[ 2] = 16'h1234; // ADDR
  spi_data[ 3] = 16'hFF00; // ADDR
  spi_data[ 4] = 16'h0000; // SIZE
  spi_data[ 5] = 16'd32; // SIZE
  
  spi_data[ 6] = 16'hAAAA;
  spi_data[ 7] = 16'h1010;
  spi_data[ 8] = 16'hAAAA;
  spi_data[ 9] = 16'h1010;
  spi_data[10] = 16'hAAAA;
  spi_data[11] = 16'h1010;
  spi_data[12] = 16'hAAAA;
  spi_data[13] = 16'h1010;
  spi_data[14] = 16'hAAAA;
  spi_data[15] = 16'h2222;

  spi_data[16] = MAGIC_WORD;
  spi_data[17] = MAGIC_CPU_UNRST;
  spi_data[18] = 16'd0;
  spi_data[19] = 16'd0;
  spi_data[20] = 16'd0;
  spi_data[21] = 16'd0;
end

enum {RST, LOAD, SEND, DELAY, FINISH} state, next;

always_ff @(posedge clk)
  state <= (rst)? RST : next;

reg [31:0] delay_cnt;
reg [31:0] ram_addr;
reg [31:0] bit_cnt;
reg [DATA_W-1:0] send_word;
always_comb begin
  next = state;
  case(state)
         RST: next = LOAD;
        LOAD: next = SEND;
        SEND: if (bit_cnt == DATA_W * 2 - 1) next = DELAY;
       DELAY: if (delay_cnt == 0)
                next = (ram_addr == SPI_WORDS - 1)? FINISH : LOAD;
      FINISH: next = FINISH;
      default: next = RST;
  endcase
end

always_ff @(posedge clk) 
  delay_cnt <= (state == DELAY) ? delay_cnt - 1 : (ram_addr == SPI_WORDS_MAGIC - 1)? MAGIC_TO_DATA_DELAY : WORDS_DELAY;

always_ff @(posedge clk)
  bit_cnt <= (state == SEND)? bit_cnt + 1 : 'd0;

always_ff @(posedge clk)
  if (state == RST)
    ram_addr <= 0;
  else if (state == DELAY && next == LOAD)
    ram_addr <= ram_addr + 1;

always_ff @(posedge clk)
  if (state == LOAD)
    send_word <= spi_data[ram_addr];

reg _spi_sck;
always_ff @(posedge clk) begin
  _spi_sck <= (state == SEND) ? ~_spi_sck : 1'b1;
  //spi_sck <= _spi_sck;
end

assign spi_sck = _spi_sck;

reg _spi_sdi;
always_ff @(posedge clk) begin
   spi_sdo  <= ((state == SEND) && ~spi_cs)            ? send_word[DATA_W - bit_cnt[31:1] - 1] : 1'b1;
  spi_sdo_d <= ((state == SEND) && ~spi_sdi_as_output) ? send_word[DATA_W - bit_cnt[31:1] - 1] : 1'b1;
end

//assign spi_sdi = (~spi_sdi_as_output)? _spi_sdi : 1'bz;

always_ff @(posedge clk) begin
  if (state == RST) begin
    spi_cs <= 1'b1;
    spi_sdi_as_output <= 1'b1;
  end
  else if (state == LOAD && ram_addr == 'd0) begin
    spi_cs <= 1'b0;
  end
  else if (state == DELAY && ram_addr == SPI_WORDS_MAGIC - 1) begin
    spi_cs <= 1'b1;
  end
  else if (state == LOAD && ram_addr >= SPI_WORDS_MAGIC) begin
    spi_sdi_as_output <= 1'b0;
  end
  else if (state == FINISH) begin
    spi_sdi_as_output <= 1'b1;
    spi_cs <= 1'b1;
  end
end

// RX
reg [15:1] spi_data_shiftreg;
reg [15:0] spi_rx_data;
reg [ 3:0] spi_bit_cnt;
always @(posedge spi_sck)
  spi_data_shiftreg <= {spi_data_shiftreg[14:1], spi_sdi};

always @(posedge spi_sck or posedge spi_cs)
	spi_bit_cnt <= (~spi_cs | ~spi_sdi_as_output) ? spi_bit_cnt + 1 : 'd0;

wire spi_bit_15 = (spi_bit_cnt == 15) ? 1 : 0;

always @(posedge spi_sck)
	if (spi_bit_15) 
		spi_rx_data <= {spi_data_shiftreg[15:1], spi_sdi};

endmodule
